//-----------------------------------------------------------------------------
//--- Texture Advection for Unsteady Flow Visualization
//--- Bruno Jobard - 2000
//--- CSIT -  Florida State University
//-----------------------------------------------------------------------------

#include "lea_advection.h"

#ifdef WIN32
double
drand48()
{
	return ((double)rand())/(double)RAND_MAX;
}

#endif

#define random rand

Advection::Advection(int size_X, int size_Y, int h_max)
{
  sizeX = size_X;
  sizeY = size_Y;
  size  = sizeX*sizeY;
  hmax  = h_max;
  h     = (float)hmax;

  border = (int)(ceilf((float)hmax));
  dimx   = sizeX + 2*border;
  dimy   = sizeY + 2*border;
  dim    = dimx*dimy;

  border = hmax;              // border size around arrays = [h]+1
  mini = border;
  maxi = dimx - border;
  minj = border;
  maxj = dimy - border;

  swapPercent = 2.0;
  blendCoef1  = (float)0.1;
  
  cx1 = new float[size];
  cy1 = new float[size];
  cx2 = new float[size];
  cy2 = new float[size];
  n1  = new float[dim];
  n2  = new float[dim];
  nb = new float[size];
  n3 = new float[size];
  nc  = n1;
  na  = n2;
  nf  = nb;
  
  int i,j,j1,ind1,ind2;
    
  //--- Init Coords
  for(j=0; j<sizeY; j++)
    {
    j1 = j*sizeX;
    for(i=0; i<sizeX; i++)
      {
      ind1 = j1+i;
      cx1[ind1] = (float)(i+drand48());
      cy1[ind1] = (float)(j+drand48());
      }
    }

  //--- Init Noise
  for(i=0; i<dim; i++)
    {
    nc[i] = (float)(drand48()<0.5 ? 0.0 : 1.0);
    }
  for(j=0; j<sizeY; j++)
    {
    j1=j*sizeX;
    for(i=0; i<sizeX; i++)
      {
      ind1=j1+i;
      ind2 = (j+minj)*dimx + i+mini;
      nb[ind1] = nc[ind2];
      }
    }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
float *
Advection::Compute(const float* vxx, const float* vyy, int nx, int ny,
		   const float _h, float swapPercentage, float blendCoef)
{
  h = _h;
  swapPercent = swapPercentage;
  blendCoef1  = blendCoef;
  
  ComputeIntegration(vxx,vyy,nx,ny);
  ComputeCoordUpdate();
  ComputeNoiseUpdate();
  ComputeEdgeTreatment();
  ComputeNoiseInjection();
  ComputeNoiseBlending();
  
  // Swap noises
  float* ntmp;
  ntmp = nc; nc = na; na = ntmp;

  return nf;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void
Advection::ComputeIntegration(const float* vxx, const float* vyy,
			      const int nx, const int ny)
{
  int i,j,j1,ind1;
  float velx, vely;
  
  float rx = float(nx-1)/float(sizeX);
  float ry = float(ny-1)/float(sizeY);

  for(j=0; j<sizeY; j++)
    {
    j1=j*sizeX;
    for(i=0; i<sizeX; i++)
      {
      ind1=j1+i;
    
      //--- Linear interpolation of velocity for each coordinate in (cx1,cy1)
      linearInterpol2D(vxx,vyy, nx,ny, cx1[ind1]*rx,cy1[ind1]*ry, &velx,&vely);
    
      //--- Integration
      cx2[ind1] = cx1[ind1] - h*velx;
      cy2[ind1] = cy1[ind1] - h*vely;
      }
    } 
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void
Advection::ComputeCoordUpdate()
{
  int i,j,j1,ind1;    
  //--- Coordinates updates
  for(j=0; j<sizeY; j++)
    {
    j1=j*sizeX;
    for(i=0; i<sizeX; i++)
      {
      ind1=j1+i;

      // Keep fractional part from x2 and y2:
      //                                       x1(i) = i + frac(x2(i))
      //                                       y1(j) = j + frac(x2(j))
      cx1[ind1]=float(i)+cx2[ind1]-int(cx2[ind1]);
      cy1[ind1]=float(j)+cy2[ind1]-int(cy2[ind1]);
      }
    }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void
Advection::ComputeNoiseUpdate()
{
  int i,j,j1,ind1,ind2,ind3;
  float nval;

  //--- Noise update

  for(j=0; j<sizeY; j++)
    {
    j1=j*sizeX;
    for(i=0; i<sizeX; i++)
      {
      ind1 = j1+i;
      ind2 = (j+minj)*dimx + i+mini;
      ind3 = int(cy2[ind1]+minj)*dimx + int(cx2[ind1]+mini);

      // constant interpolation advection of current noise
      na[ind2] = nc[ind3];

      // linear interpolation advection for final display
      linearInterpol2D(nc, dimx,dimy, float(cx2[ind1]+mini),float(cy2[ind1]+minj), &nval);
      n3[ind1] = nval;
      }
    }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void
Advection::ComputeEdgeTreatment()
{ 
  int i,j,j1,ind1;
  
  for(j=0; j<minj; j++)
    {
    j1=j*dimx;
    for(i=0; i<dimx; i++)
      {
      ind1=j1+i;
      na[ind1] = float(random()&01);
      }
    }
  for(j=maxj; j<dimy; j++)
    {
    j1=j*dimx;
    for(i=0; i<dimx; i++)
      {
      ind1=j1+i;
      na[ind1] = float(random()&01);
      }
    }
  for(j=minj; j<maxj; j++)
    {
    j1=j*dimx;
    for(i=0; i<mini; i++)
      {
      ind1=j1+i;
      na[ind1] = float(random()&01);
      }
    for(i=maxi; i<dimx; i++)
      {
      ind1=j1+i;
      na[ind1] = float(random()&01);
      }
    }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void
Advection::ComputeNoiseInjection()
{
  int i,ind1;
  //--- Noise injection
  int incr = int(100/swapPercent);

  for(i=minj*dimx; i<maxj*dimx-incr; i+=incr)
    {
    ind1 = int(i+drand48()*incr);     // swap one pixel in each incr interval

    //int isbefore = 0;
    //if (na[ind1]>1.0 || na[ind1]<0.0) isbefore = 1;
    na[ind1] = (float)(1.0 - na[ind1]); 
    }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void
Advection::ComputeNoiseBlending()
{
  int i,j,j1,ind1;
  for(j=0; j<sizeY; j++)
    {
    j1=j*sizeX;
    for(i=0; i<sizeX; i++)
      {
      ind1 = j1+i;
      nb[ind1] += blendCoef1*(n3[ind1]-nb[ind1]);
      }
    }
  nf = nb;   // Current final noise is the blended version
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void
WritePGM(char *filename, int width, int height, float* image)
{
  FILE *pgmfile;

  pgmfile = fopen(filename, "wb");

  //--- header
  fprintf(pgmfile, "P5\n%d %d 255\n", width, height);

  //--- image
  fwrite(image, sizeof(unsigned char), width*height, pgmfile);

  fclose(pgmfile);
}
