#include "vtkLagrangianEulerianAdvection.h"

#include "vtkFloatArray.h"
#include "vtkImageData.h"
#include "vtkUnsignedCharArray.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
 
#include <math.h>
#include "lea_advection_master.h"

vtkCxxRevisionMacro(vtkLagrangianEulerianAdvection, "$Revision: 1.1 $");
vtkStandardNewMacro(vtkLagrangianEulerianAdvection);

// Construct object 

vtkLagrangianEulerianAdvection::vtkLagrangianEulerianAdvection()
{
  this->XSize = 128;
  this->YSize = 128;
  this->IntegrationStep = 5;
  this->BlendCoefficient = 0.1;

  this->EnableLIC = 1;
  this->NormalizationLIC = 1;
  this->IntensityCorrectionLIC = 0;
  this->KernelLengthLIC = 8;
  this->KernelShape = 1;
  this->SwapPercentage = 3;
  this->VelocityMask = 0;
  this->VelocityTransparency = 0;
  this->VelocityMaskContrast = 2;

  this->InitialSteps = 10;
  this->StepsAtOnce = 1;

  this->Run = 0;
  this->AspectRatio = 1;
  this->lea = new AdvectionMaster();

  this->VelocityXComponent = NULL;
  this->VelocityYComponent = NULL;
  this->VelocityMagnitude = NULL;
  this->VelocityMax = 0;
}


vtkLagrangianEulerianAdvection::~vtkLagrangianEulerianAdvection()
{
  if (this->lea)
    {
    delete this->lea;
    }
  if(this->VelocityXComponent)
    delete [] this->VelocityXComponent;
  if(this->VelocityYComponent)
    delete [] this->VelocityYComponent;
  if(this->VelocityMagnitude)
    delete [] this->VelocityMagnitude;
}

void vtkLagrangianEulerianAdvection::SetXSize(int nx)
{
  if(this->XSize != nx)
    {
    this->XSize = nx;
    this->YSize = (int)(nx / this->AspectRatio);
    this->Modified();
    }
}

void vtkLagrangianEulerianAdvection::SetYSize(int ny)
{
  if(this->YSize != ny)
    {
    this->YSize = ny;
    this->XSize = (int)(ny * this->AspectRatio);
    this->Modified();
    }
}

void vtkLagrangianEulerianAdvection::ExecuteInformation()
{
  double bounds[6], spacing[3]; 
  int dims[6];

  vtkImageData *output = this->GetOutput();
  vtkImageData *input = static_cast<vtkImageData *>(this->GetInput());

  // Set volume origin and data spacing
  input->GetBounds(bounds);
  input->GetSpacing(spacing);
  input->GetWholeExtent(dims);

  output->SetOrigin(input->GetOrigin());
  
  output->SetSpacing((bounds[1]-bounds[0])/(this->XSize-1),
                     (bounds[3]-bounds[2])/(this->YSize-1),
                     spacing[2]);

  output->SetWholeExtent(0, this->XSize-1, 
                         0, this->YSize-1, 
                         dims[4], dims[5]);

  output->SetScalarType(VTK_UNSIGNED_CHAR);
  output->SetNumberOfScalarComponents(3);

  this->AspectRatio = (dims[1]+1.0) / (dims[3]+1.0);
  vtkDebugMacro( << "AspectRatio = " << this->AspectRatio);
  this->NumberOfNodes = (dims[1]+1) * (dims[3]+1);
  if(this->VelocityXComponent == NULL)
    {
    this->VelocityXComponent = new float[this->NumberOfNodes];
    this->VelocityYComponent = new float[this->NumberOfNodes];
    this->VelocityMagnitude  = new float[this->NumberOfNodes];
    }

  this->lea->SetResolution(this->XSize, this->YSize, VTK_LEA_INTEGRATION_STEP_MAX);

  this->lea->SetAdvectionOptions(this->IntegrationStep,
                                 this->SwapPercentage,
                                 this->BlendCoefficient);
  this->lea->SetLicOptions(this->KernelLengthLIC,
                           this->KernelShape,
                           this->IntensityCorrectionLIC,
                           this->NormalizationLIC);

  this->lea->SetVelocityMaskOn(VelocityMask);
  this->lea->SetVelocityMaskOptions(this->VelocityMaskContrast,
                                    this->VelocityTransparency);
}

void vtkLagrangianEulerianAdvection::ExecuteData(vtkDataObject *outp)
{
  int i, dims[6];
  float color[4];
  vtkImageData *output = this->AllocateOutputData(outp);

  vtkImageData *input = static_cast<vtkImageData *>(this->GetInput());

  input->GetWholeExtent(dims);

  vtkFloatArray *VectorField = (vtkFloatArray *) input->GetPointData()->GetVectors();

  if(VectorField == NULL) // did not find a vector, try a scalar with size=3
    {
    VectorField = (vtkFloatArray *)input->GetPointData()->GetScalars();
    if(VectorField->GetNumberOfComponents() != 3)
      {
      return;
      }
    }
  else
    {
    if(this->Run == 0)
      {
      float *vfdata = VectorField->GetPointer(0);
      float magmax;
      float valvx, valvy, valmag;

      if(this->VelocityMax > 0)
        {
        magmax = this->VelocityMax;
        }
      else
        {
        magmax = VectorField->GetMaxNorm();
        }
      for (i=0; i< this->NumberOfNodes; i++)
        {
        valvx = vfdata[i*3  ];
        valvy = vfdata[i*3+1];
        valmag= sqrt(valvx*valvx + valvy*valvy);
  
        if (valmag >= magmax)
          {
          this->VelocityXComponent[i]   = valvx / valmag;
          this->VelocityYComponent[i]   = valvy / valmag;
          this->VelocityMagnitude[i] = 1;
          }
        else
          {
          this->VelocityXComponent[i]   = valvx / magmax;
          this->VelocityYComponent[i]   = valvy / magmax;
          this->VelocityMagnitude[i] = valmag/ magmax;
          }
        }
  
      this->lea->SetCurrentVectorField(this->VelocityXComponent, this->VelocityYComponent, this->VelocityMagnitude, (dims[1]+1) , (dims[3]+1));
      }
    if (this->InitialSteps && this->InitialSteps>0)
      {
      this->lea->SetLicOn(0);
      this->lea->SetVelocityMaskOn(0);
      for (int i=0; i<this->InitialSteps; i++)
        {
        this->lea->Compute();
        }
      this->InitialSteps = 0;
      }
    else if (this->StepsAtOnce > 1)
      {
      // Run internal advection nStepsAtOnce times
      this->lea->SetLicOn(0);
      this->lea->SetVelocityMaskOn(0);
      for (int i=0; i< this->StepsAtOnce-1; i++)
        {
        this->lea->Compute();
        }
      }
    // Run internal advection a last time with correct options
    this->lea->SetLicOn(EnableLIC);
    this->lea->SetVelocityMaskOn(this->VelocityMask);
    this->lea->Compute();

    vtkDataArray *RGB = output->GetPointData()->GetScalars();
    RGB->SetName("rgb");

    int nbComponents;
    float *src_img = lea->GetOutput(nbComponents);

    for(i=0; i < this->XSize * this->YSize; i++)
      {
      color[0] = color[1] = color[2] = src_img[i]*255.0;
      RGB->SetTuple(i, color);
      }
    }
}

void vtkLagrangianEulerianAdvection::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
  os << indent << "XSize: " << this->XSize << "\n";
  os << indent << "YSize: " << this->YSize << "\n";
  os << indent << "IntegrationStep: " << this->IntegrationStep << "\n";
  os << indent << "BlendCoefficient: " << this->BlendCoefficient << "\n";
  os << indent << "EnableLIC: " << this->EnableLIC << "\n";
  os << indent << "NormalizationLIC: " << this->NormalizationLIC << "\n";
  os << indent << "IntensityCorrectionLIC: " << this->IntensityCorrectionLIC << "\n";
  os << indent << "KernelLengthLIC: " << this->KernelLengthLIC << "\n";
  os << indent << "VelocityMask: " << this->VelocityMask << "\n";
  os << indent << "VelocityTransparency: " << this->VelocityTransparency << "\n";
  os << indent << "VelocityMaskContrast: " << this->VelocityMaskContrast << "\n";
  os << indent << "InitialSteps: " << this->InitialSteps << "\n";
  os << indent << "StepsAtOnce: " << this->StepsAtOnce << "\n";
  os << indent << "Run: " << this->Run << "\n";
}
