// .NAME vtkLagrangianEulerianAdvection - creates 2D unsteady flow textures
// .SECTION Description
// vtkLagrangianEulerianAdvection is a filter that transforms 2D
// transient vector fields as a series of spatially and temporally correlated
// bitmap images. The input to vtkLagrangianEulerianAdvection can be either
// a single two-dimensional vtkStructuredData with a VECTORS attribute, or it
// can be a series of time-dependent vtkStructuredData. The output is a
// vtkImageData.
//
// The algorithm used is the one described by Bruno Jobard, Gordon Erlebacher
// and Youssuf Hussaini in their Visualization 2001 paper, "Lagrangian-Eulerian
// Advection for Unsteady Flow Visualization". The core of this VTK class is the
// original algorithm implementation provided by the authors. The general
// principle of the algorithm is to perform a per-pixel advection of an internal
// noise texture along the flow to achieve a high temporal correlation and to
// apply a temporal filter onto successive textures in order to introduce
// spatial correlation. A single resulting image shows the structure of the
// input vector field. Giving the filter subsequent time steps of a vector field
// will produce a series of output images that will depict the transient
// evolution of the vector field.
// .SECTION Caveats
// 
//
// .SECTION Thanks
// Thanks to Jean M. Favre of the Swiss Center for Scientific Computing (CSCS)
// for creating the VTK class and to Bruno Jobard (University of Pau, France)
// for guidance during the port.
//


#ifndef __vtkLEA_h
#define __vtkLEA_h

#include "vtkDataSetToImageFilter.h"
#include "vtkPVLEAConfigure.h"

//BTX
class AdvectionMaster;
//ETX

#define VTK_LEA_INTEGRATION_STEP_MAX 20

class VTK_PVLEA_EXPORT vtkLagrangianEulerianAdvection : public vtkDataSetToImageFilter 
{
public:
  vtkTypeRevisionMacro(vtkLagrangianEulerianAdvection,vtkDataSetToImageFilter);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Construct object with texture resolution=(128,128)
  static vtkLagrangianEulerianAdvection *New();
  
  // Description:
  // Set the X texture map dimension. Default is 128.
  void SetXSize(int);
  vtkGetMacro(XSize, int);

  // Description:
  // Set the Y texture map dimension. Default is 128.
  void SetYSize(int);
  vtkGetMacro(YSize, int);

  // Description:
  // Set a Integration Step
  vtkGetMacro(IntegrationStep, int);
  vtkSetClampMacro(IntegrationStep, int, 1, VTK_LEA_INTEGRATION_STEP_MAX);

  // Description:
  // Set a Blend Coefficient
  vtkGetMacro(BlendCoefficient, float);
  vtkSetClampMacro(BlendCoefficient, float, 0.05, 1.0);

  // Description:
  // Get a SwapPercentage
  vtkGetMacro(SwapPercentage, float);
  
  // Description:
  // Set a EnableLIC
  vtkGetMacro(EnableLIC, int);
  vtkSetClampMacro(EnableLIC, int, 0, 1);

  // Description:
  // Set a NormalizationLIC
  vtkGetMacro(NormalizationLIC, int);
  vtkSetClampMacro(NormalizationLIC, int, 0, 1);

  // Description:
  // Set a IntensityCorrectionLIC
  vtkGetMacro(IntensityCorrectionLIC, int);
  vtkSetClampMacro(IntensityCorrectionLIC, int, 0, 1);

  // Description:
  // Set a KernelLengthLIC
  vtkGetMacro(KernelLengthLIC, int);
  vtkSetClampMacro(KernelLengthLIC, int, 2, 40);

  // Description:
  // Get a KernelShape
  vtkGetMacro(KernelShape, int);

  // Description:
  // Set a VelocityMask
  vtkGetMacro(VelocityMask, int);
  vtkSetClampMacro(VelocityMask, int, 0, 1);

  // Description:
  // Set a VelocityTransparency
  vtkGetMacro(VelocityTransparency, int);
  vtkSetClampMacro(VelocityTransparency, int, 0, 1);

  // Description:
  // Set a VelocityMaskContrast
  vtkGetMacro(VelocityMaskContrast, int);
  vtkSetClampMacro(VelocityMaskContrast, int, 1, 3);

  // Description:
  // Set a InitialSteps
  vtkGetMacro(InitialSteps, int);
  vtkSetClampMacro(InitialSteps, int, 1, 30);

  // Description:
  // Set a StepsAtOnce
  vtkGetMacro(StepsAtOnce, int);
  vtkSetClampMacro(StepsAtOnce, int, 1, 10);

  // Description:
  // Set a VelocityMax
  vtkGetMacro(VelocityMax, float);
  vtkSetClampMacro(VelocityMax, float, 0, 100);

  // Description:
  // Set Run On and Off. Used for texture animation with a single input vector
  vtkGetMacro(Run, int);
  vtkSetClampMacro(Run, int, 0, 1);
  void RunOn() {this->SetRun(1);};
  void RunOff() {this->SetRun(0);};

protected:
  vtkLagrangianEulerianAdvection();
  ~vtkLagrangianEulerianAdvection();



  void ExecuteInformation();
  void ExecuteData(vtkDataObject *);

  int XSize;
  int YSize;

  int IntegrationStep;
  float BlendCoefficient;
  float SwapPercentage;

  int EnableLIC;
  int NormalizationLIC;
  int IntensityCorrectionLIC;
  int KernelLengthLIC;
  int KernelShape;

  int VelocityMask;
  int VelocityTransparency;
  int VelocityMaskContrast;

  int InitialSteps;
  int StepsAtOnce;

  int Run;

  int NumberOfNodes;
  float AspectRatio;
  float *VelocityXComponent;
  float *VelocityYComponent;
  float *VelocityMagnitude;
  float VelocityMax;
private:
  AdvectionMaster* lea;

  vtkLagrangianEulerianAdvection(const vtkLagrangianEulerianAdvection&);  // Not implemented.
  void operator=(const vtkLagrangianEulerianAdvection&);  // Not implemented.
};

#endif
